#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from io import BytesIO, SEEK_CUR, SEEK_END, SEEK_SET
from pathlib import Path

import numpy as np
from PIL import Image

from . import NVDecoder, YVDecoder
from .FloDecoder import open_flo, KITTI

Image.register_decoder('NV12', NVDecoder.NV12Decoder)
Image.register_decoder('NV21', NVDecoder.NV21Decoder)
Image.register_decoder('YV12', YVDecoder.YV12Decoder)
Image.register_decoder('YV21', YVDecoder.YV21Decoder)


class File:
  """An abstract file object

  NOTE: If `path` is a file, `File` opens it and calculates its length.
  If `path` is a folder, `File` opens every file in the folder in abc order.

  Args:
       path: path to a **node**, where node can be a **file** or a **folder**
         contains multiple files.
       rewind: rewind the file automatically when reaches EOF.
  """

  def __init__(self, path, rewind=False):
    self.path = Path(path)
    self.file = []
    self.length = dict()
    self.name = self.path.stem
    self.full_name = self.path.absolute().as_posix()
    if self.path.is_file():
      self.file = [self.path]
      self.length[self.path.name] = self.path.stat().st_size
    elif self.path.is_dir():
      for _file in self.path.glob('*'):
        self.file.append(_file)
        self.length[_file.name] = _file.stat().st_size
      # sort the files by name, because they are unordered in UNIX
      self.file.sort()
    self.file_ = self.file.copy()
    self.read_file = []
    self.read_pointer = 0
    self.end_pointer = sum(self.length.values())
    self.cur_fd = None
    self.rewind = rewind

  def __len__(self):
    """total size of file or files"""
    return self.end_pointer

  def _seek(self, target):
    """seek to a target position of `File`

    Args:
        target: an int representing byte position.
    """
    assert 0 <= target < self.end_pointer
    if self.read_pointer == target:
      # do not need operation
      return
    if not self.cur_fd and self.file:
      # guarantee an opened fd, happens if `File` is just initialized.
      self.cur_fd = self.file[0].open('rb')
      self.read_file.append(self.file.pop(0))
    while self.read_pointer < target:
      # move forward
      reminder = self.length.get(self.read_file[-1].name) - \
                 self.cur_fd.tell()
      if self.read_pointer + reminder >= target:
        self.cur_fd.seek(target - self.read_pointer, SEEK_CUR)
        self.read_pointer = target
        return
      else:
        self.read_pointer += reminder
        self.cur_fd.close()
        self.cur_fd = self.file[0].open('rb')
        self.read_file.append(self.file.pop(0))
    while self.read_pointer > target:
      # move backward
      reminder = self.cur_fd.tell()
      if self.read_pointer - reminder <= target:
        self.cur_fd.seek(target - self.read_pointer, SEEK_CUR)
        self.read_pointer = target
        return
      else:
        self.read_pointer -= reminder
        self.cur_fd.close()
        self.cur_fd = self.read_file[-1].open('rb')
        self.file.insert(0, self.read_file.pop())

  def reopen(self):
    """clear the current state and re-initialize read pointer"""
    self.file = self.file_.copy()
    self.read_file.clear()
    self.read_pointer = 0
    self.cur_fd = None

  def split(self, depth):
    raise NotImplementedError

  def read(self, count=None):
    """Read `count` bytes

    Args:
        count: size of bytes to read, if None (default),
          read all bytes of current file

    Return:
        bytes: bytes read
    """
    if count == 0:
      return b''
    if not self.cur_fd and self.file:
      self.cur_fd = self.file[0].open('rb')
      self.read_file.append(self.file.pop(0))
    elif not self.cur_fd:
      if self.rewind and self.read_file:
        self.reopen()
        return self.read(count)
      else:
        raise EOFError(f'End of File! {self.name}')
    read_bytes = self.cur_fd.read(count)
    if read_bytes:
      self.read_pointer += len(read_bytes)
      if count and count > len(read_bytes):
        return read_bytes + self.read(count - len(read_bytes))
      elif count:
        return read_bytes
      else:
        # read entire file, close the file descriptor
        self.cur_fd.close()
        self.cur_fd = None
        return read_bytes
    else:
      if self.file:
        self.cur_fd.close()
        self.cur_fd = self.file[0].open('rb')
        self.read_file.append(self.file.pop(0))
        return self.read(count)
      elif self.rewind and self.read_file:
        self.reopen()
        return self.read(count)
      else:
        raise EOFError(f'End of File! {self.name}')

  def read_frame(self, frames=1, *args):
    """An abstract interface"""
    raise NotImplementedError

  def seek(self, offset, where):
    """Seek the position by `offset` relative to `where`.

    Args:
         offset: move the read pointer by `offset` bytes.
         where: same as io.SEEK_END, io.SEEK_CUR or io.SEEK_SET.
    """
    if where == SEEK_SET:
      self._seek(offset)
    if where == SEEK_CUR:
      self._seek(self.read_pointer + offset)
    if where == SEEK_END:
      self._seek(self.end_pointer + offset)

  def tell(self):
    """Tell the current position of the read pointer."""
    return self.read_pointer

  def size(self, name=None):
    """Get the length of the file named `name`

    Args:
        name: specify a named file.

    Return:
        int: length in bytes
    """
    if name is None:
      return sum(self.length.values())
    path = Path(name)
    name = path.stem if path.exists() else name
    return self.length.get(name)


# Supported RAW format, see 'FOURCC' standard
_ALLOWED_RAW_FORMAT = [
  'YV12',  # [Y][U][V]
  'YV21',  # [Y][V][U]
  'NV12',  # [Y][UV]
  'NV21',  # [Y][VU]
  'RGB',  # [RGB]
  'BGR',  # [BGR]
  'RGBA',  # [RGBA]
  'BGRA',  # [BGRA]
]


class RawFile(File):
  """For reading raw files. The file is lazy loaded, which means
  the file is opened but not loaded into memory at initialization.

  Args:
       path: a string representing `node` path.
       mode: a string, since raw file has no headers, type must be
         explicitly given, see `_ALLOWED_RAW_FORMAT`.
       size: a tuple of int (width, height). If `path` is a folder,
         all files in it must be the same shape.
       rewind: rewind the file automatically when reaches EOF

  Raise:
      TypeError: if `mode` is not supported
  """

  def __init__(self, path, mode, size, rewind=False):

    if not mode.upper() in _ALLOWED_RAW_FORMAT:
      raise TypeError('unknown mode: ' + mode)
    self.mode = mode.upper()
    self._size = size
    self.pitch, self.channel_pitch = self._get_frame_pitch()
    super(RawFile, self).__init__(path, rewind)
    self._pair = None

  def _get_frame_pitch(self):
    """Get bytes length of one frame.
    For the detail of mode fourcc, please see https://www.fourcc.org/

    NOTE: RGB, BGR, and UV channel of NV12, NV21 is packed, while YV12 and
      YV21 is planar, hence we have:
      - **channel0** of YV12, YV21, NV12, NV21 if Y
      - **channel1** of YV12 is U, YV21 is V, NV12 is UV, NV21 is VU
      - **channel2** of YV12 is V, YV21 is U
    """
    mode = self.mode
    w, h = self._size
    if mode in ('YV12', 'YV21'):
      return h * w * 3 // 2, [h * w, h * w // 4, h * w // 4]
    if mode in ('NV12', 'NV21'):
      return h * w * 3 // 2, [h * w, h * w // 2]
    if mode in ('RGB', 'BGR'):
      return h * w * 3, [h * w * 3]
    if mode in ('RGBA', 'BGRA'):
      return h * w * 4, [h * w * 4]

  def _get_frame_channel_shape(self):
    """Get each channel's shape according to mode and frame length.
    For the detail of mode fourcc, please see https://www.fourcc.org/
    """
    mode = self.mode
    w, h = self._size
    if mode in ('YV12', 'YV21'):
      return (np.array([1, h, w]),
              np.array([1, h // 2, w // 2]),
              np.array([1, h // 2, w // 2]))
    if mode in ('NV12', 'NV21'):
      return np.array([1, h, w]), np.array([2, h // 2, w // 2])
    if mode in ('RGB', 'BGR'):
      return np.array([h, w, 3])
    if mode in ('RGBA', 'BGRA'):
      return np.array([h, w, 4])

  def read_frame(self, frames=1, *args):
    """read number of `frames` of the file. A frame is a single image

    Args:
        frames: number of frames to be loaded.
    """
    if self.mode in ('YV12', 'YV21', 'NV12', 'NV21',):
      ret = []
      for _ in range(frames):
        data = self.read(self.pitch)
        ret.append(Image.frombytes('YCbCr', self._size, data, self.mode))
      return ret
    elif self.mode in ('RGB', 'RGBA'):
      ret = []
      for _ in range(frames):
        data = self.read(self.pitch)
        ret.append(Image.frombytes(self.mode, self._size, data))
      return ret
    elif self.mode in ('BGR',):
      _image_mode = 'RGB'
      ret = []
      for _ in range(frames):
        data = b''.join(
          (self.read(3)[::-1] for _ in range(self.pitch // 3)))
        ret.append(Image.frombytes('RGB', self._size, data))
      return ret
    elif self.mode in ('BGRA',):
      ret = []
      for _ in range(frames):
        buf = bytes()
        for _ in range(self.pitch // 4):
          c = self.read(4)
          buf.join((c[2::-1], c[3:]))
        ret.append(Image.frombytes('RGBA', self._size, buf))
      return ret

  def seek(self, offset, where=SEEK_SET):
    """Seek the position by `offset` relative to `where`.

    Args:
         offset: move the read pointer by `offset` bytes.
         where: same as io.SEEK_END, io.SEEK_CUR or io.SEEK_SET.
    """
    if where == SEEK_SET:
      super(RawFile, self).seek(offset * self.pitch, where)
    if where == SEEK_CUR:
      super(RawFile, self).seek(
        offset * self.pitch - self.tell() % self.pitch, where)
    if where == SEEK_END:
      super(RawFile, self).seek(offset * self.pitch, where)

  def pad(self, padding):
    """RawFile doesn't support pad for now"""

    print(" [!] warning: pad is not supported in RawFile")

  def attach_pair(self, pair_file):
    self._pair = RawFile(pair_file, self.mode, self._size, self.rewind)
    return self

  @property
  def pair(self):
    return self._pair

  @property
  def shape(self):
    return self._size

  @property
  def frames(self):
    """frames in `RawFile`"""
    return self.end_pointer // self.pitch


class ImageFile(File):
  """Open image file or a sequence of image frames

  Args:
      path: a string representing `node` path.
      rewind: rewind the file when reaches EOF.
  """

  def __init__(self, path, rewind=False):
    super(ImageFile, self).__init__(path, rewind)
    self._flow = None
    self._pair = None

  def read_frame(self, frames=1, *args):
    """read number `frames` of the file.

    Args:
        frames: number of frames to be loaded
    """
    image_bytes = [BytesIO(self.read()) for _ in range(frames)]
    return [Image.open(fp) for fp in image_bytes]

  def read_frame2(self, frames=1, *args):
    """new API, saving memory while loading frames. But will consume a lot of
    file descriptors.

    Args:
        frames: number of frames to be loaded
    """
    imgs = []
    if frames == 0:
      return imgs
    while True:
      if len(self.file) > 0:
        cur_fd = self.file.pop(0)
        imgs.append(Image.open(cur_fd))
        self.read_file.append(cur_fd)
        with open(cur_fd, 'rb') as fd:
          fd.seek(0, SEEK_END)
          self.read_pointer += fd.tell()
      elif self.rewind:
        self.reopen()
      else:
        raise EOFError('End of File!')
      if len(imgs) == frames:
        break
    return imgs

  def seek(self, offset, where=SEEK_SET):
    """Seek the position by `offset` relative to `where`.

    Args:
        offset: move the read pointer by `offset` bytes.
        where: same as io.SEEK_END, io.SEEK_CUR or io.SEEK_SET.
    """
    if where == SEEK_CUR:
      cur = len(self.read_file)
      pos = cur + offset
    elif where == SEEK_END:
      pos = len(self.read_file) + len(self.file) + offset
    else:
      pos = offset
    if pos < 0:
      pos = 0
    self.file = self.read_file + self.file
    self.read_file = self.file[:pos]
    self.file = self.file[pos:]
    self.cur_fd = None

  def pad(self, padding):
    """Pad file(s) list in the head and tail.
      Padded file is temperate and will be dropped after `reopen()`

    Args:
      padding: a integer or a list of 2 integers.
    """

    if not isinstance(padding, (list, tuple)):
      padding = [padding, padding]
    else:
      assert len(padding) is 2
    if self.read_file:
      raise RuntimeError(
        "pad must be called when reading cursor is at the beginning.")
    for _ in range(padding[0]):
      self.file.insert(0, self.file[0])
    for _ in range(padding[1]):
      self.file.append(self.file[-1])

  def attach_flow(self, flow_file):
    self._flow = flow_file
    return self

  @property
  def flow(self):
    fd = Path(self._flow)
    assert fd.exists()
    if fd.suffix == '.flo':
      flow = open_flo(str(fd))
    elif fd.suffix == '.png':
      flow = KITTI.open_flow(str(fd))
    else:
      raise TypeError('unsupported flow format', fd.suffix)
    return flow

  def attach_pair(self, pair_file):
    self._pair = ImageFile(pair_file, self.rewind)
    return self

  @property
  def pair(self):
    return self._pair

  @property
  def shape(self):
    if self.file:
      file = self.file[0]
    else:
      file = self.read_file[0]
    with Image.open(file) as img:
      return img.width, img.height

  @property
  def frames(self):
    return len(self.file) + len(self.read_file)



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7
#  Collect and reproduce Image/Video Super-Resolution Algorithms.

from __future__ import absolute_import
from __future__ import print_function

import VSR.Backend
import VSR.DataLoader

__all__ = [
  'Backend',
  'DataLoader',
  'Model',
  'Util',
]



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import logging
from concurrent import futures

import numpy as np
from PIL import Image
from psutil import virtual_memory

from .Crop import RandomCrop
from .Dataset import Container, Dataset
from .Transform import Bicubic, Tidy
from ..Backend import DATA_FORMAT
from ..Util import Utility
from ..Util.ImageProcess import img_to_array

FREE_MEMORY = virtual_memory().available * 0.5
LOG = logging.getLogger('VSR.Loader')


def _augment(image, op):
  """Image augmentation"""
  assert image.ndim == 4
  if op[0]:
    if DATA_FORMAT == 'channels_last':
      image = np.rot90(image, 1, axes=(1, 2))
    else:
      image = np.rot90(image, 1, axes=(2, 3))
  if op[1]:
    if DATA_FORMAT == 'channels_last':
      image = image[:, :, ::-1]
    else:
      image = image[..., ::-1]
  if op[2]:
    if DATA_FORMAT == 'channels_last':
      image = image[:, ::-1]
    else:
      image = image[:, :, ::-1]
  return image


class EpochIterator:
  """An iterator for generating batch data in one epoch

  Args:
      loader: A `Loader` object to provide properties.
      shape: The shape of the generated batch, 5-D requested [N, T, C, H, W].
      steps: The number of batches to generate in one epoch.
      shuffle: A boolean representing whether to shuffle the dataset.

  Note:
      The rules for -1 shape:
      - If shape[1] is -1, which represents the temporal length,
        will generate the entire video clips;
      - if shape[2] is -1, will auto choose the channel number according to
        color format;
      - if shape[3] or shape[4] is -1, will deduce the frame height or width
        to replace the value.

      The rules for steps:
      - If the `steps` is -1, will generate batches in sequential order;
  """

  def __init__(self, loader, shape, steps, shuffle=None):
    self.loader = loader
    self.shape = shape
    if steps <= 0:
      t = len(self.loader.data['hr'])
      b = self.shape[0]
      self.steps = t // b + int(np.ceil((t % b) / b))
    else:
      self.steps = steps
    self.count = 0
    if shuffle:
      mmax = len(self.loader.data['hr'])
      if mmax > 0:
        self.index = np.random.randint(mmax, size=self.steps * shape[0])
      else:
        self.steps = 0
    else:
      self.index = np.arange(self.steps * shape[0])

  def __len__(self):
    return self.steps

  def __iter__(self):
    return self

  def __next__(self):
    pack = {'hr': [], 'lr': [], 'name': []}
    if self.count >= self.steps:
      raise StopIteration("All batch data generated.")

    slc = slice(self.count * self.shape[0], (self.count + 1) * self.shape[0])
    crop = self.loader.crop
    cb_hr = (self.loader.hr['transform1'], self.loader.hr['transform2'])
    cb_lr = (self.loader.lr['transform1'], self.loader.lr['transform2'])
    for i in self.index[slc]:
      if i >= len(self.loader.data['hr']):
        continue
      hr = self.loader.data['hr'][i]
      lr = self.loader.data['lr'][i]
      name = self.loader.data['names'][i]
      hr2 = hr
      for fn in cb_hr[0]:
        hr2 = [fn(img) for img in hr2]
      hr2 = [img.convert(self.loader.hr['color']) for img in hr2]
      lr2 = lr
      for fn in cb_lr[0]:
        lr2 = [fn(img) for img in lr2]
      lr2 = [img.convert(self.loader.lr['color']) for img in lr2]
      hr3 = np.stack([img_to_array(img, DATA_FORMAT) for img in hr2])
      lr3 = np.stack([img_to_array(img, DATA_FORMAT) for img in lr2])
      del hr2, lr2
      if hr3.shape[0] == 1 and lr3.shape[0] == 1:
        hr3 = hr3.squeeze(0)
        lr3 = lr3.squeeze(0)
        hr4, lr4 = crop((hr3, lr3), shape=self.shape[2:]) if crop else (
          hr3, lr3)
      else:
        hr4, lr4 = crop((hr3, lr3), shape=self.shape[1:]) if crop else (
          hr3, lr3)
      del hr3, lr3
      hr4 = np.expand_dims(hr4, 0)  # 4-D or 5-D
      lr4 = np.expand_dims(lr4, 0)  # [1, (T,) C, H, W]
      for fn in cb_hr[1]:
        hr4 = fn(hr4)
      for fn in cb_lr[1]:
        lr4 = fn(lr4)

      if self.loader.aux['augmentation']:
        ops = np.random.randint(0, 2, [3])
      else:
        ops = [0, 0, 0]
      _shape0 = hr4.shape
      _shape1 = lr4.shape
      hr5 = _augment(hr4.reshape([-1, *_shape0[-3:]]), ops)
      lr5 = _augment(lr4.reshape([-1, *_shape1[-3:]]), ops)
      del hr4, lr4
      pack['hr'].append(hr5.reshape(_shape0))
      pack['lr'].append(lr5.reshape(_shape1))
      pack['name'].append(name)

    if pack['hr']:
      pack['hr'] = np.concatenate(pack['hr'])
    if pack['lr']:
      pack['lr'] = np.concatenate(pack['lr'])
    self.count += 1
    return pack


class Loader(object):
  """A parallel data loader that generates label and data batches each epoch.

  Args:
      hr_data: this is a data container from `Dataset` object, represents the
               label (high-resolution) data.
      lr_data: this is a data container from `Dataset` object, represents the
               training (low-resolution) data.
      scale: specify the scale factor for this model. If scale is not specified,
             you must set "lr_data", and "cropper" explicitly.
      extra_data: a dict object contains extra information. You won't use this.
      threads: num of threads used to load the data.

  Note:
      A `Loader` object has several attributes to enhance the loader's ability.
      See `Loader.add_data_transform`, `Loader.image_augmentation`,
      `Loader.cropper`, `Loader.set_color_space` for details.
  """

  def __init__(self, hr_data, lr_data=None, scale=None, extra_data: dict = None,
               threads=1):
    # check type
    if isinstance(hr_data, Dataset):
      hr_data = hr_data.compile()
      assert isinstance(hr_data, Container)
    if isinstance(lr_data, Dataset):
      lr_data = lr_data.compile()
      assert isinstance(lr_data, Container)
    if lr_data is not None and hr_data is not None:
      assert len(hr_data) == len(lr_data)
    else:
      hr_data = hr_data or lr_data
      lr_data = lr_data or hr_data
    if hr_data is None and lr_data is None:
      hr_data = lr_data = Container([], False)
    scale = scale or 1  # deduce to 1
    if extra_data is not None:
      assert isinstance(extra_data, dict)

    # default params
    self.hr = {
      'data': hr_data,
      'transform1': [],
      'transform2': [],
      'color': 'RGB'
    }
    self.lr = {
      'data': lr_data,
      'transform1': [],
      'transform2': [],
      'color': 'RGB'
    }
    self.aux = {
      'augmentation': False,
      'scale': scale,
      'fetchList': list(np.arange(len(hr_data)))
    }
    self.data = {
      'hr': [],
      'lr': [],
      'names': [],
      'extra': []
    }
    self.cache = {
      'hr': [],
      'lr': [],
      'names': [],
      'extra': []
    }
    self.extra = extra_data or {}
    self.crop = None
    self.threads = threads
    self.thp = futures.ThreadPoolExecutor(max_workers=threads)
    self.fs = []
    self.loaded = 0
    if self.hr['data'] is self.lr['data']:
      cap = self.hr['data'].capacity
    else:
      cap = self.hr['data'].capacity + self.lr['data'].capacity
    if self.extra and isinstance(self.extra['data'], Container):
      cap += self.extra['data'].capacity
    self.aux['cap'] = cap  # estimated memory usage in bytes
    if hr_data is lr_data and scale > 1:
      self.add_data_transform('hr', Tidy(scale))
      self.add_data_transform('lr', Tidy(scale), Bicubic(1 / scale))

  def add_data_transform(self, target: str, *fn, dtype='pillow'):
    """Add data transform functions. Each function will be called before
    generating batch data.

    Args:
        target: either "hr" or "lr", specify which data to apply.
        *fn: functions with only one argument, the type of the argument will
             be specified through `dtype`.
        dtype: specify the type of the function's argument.

    Note:
        `dtype` supports `numpy.ndarray` and `PIL.Image.Image`
    """
    assert target.lower() in ('hr', 'lr')
    if isinstance(dtype, Image.Image):
      dtype = 'pillow'
    elif isinstance(dtype, np.ndarray):
      dtype = 'numpy'
    assert dtype.lower() in ('pillow', 'numpy', 'pil', 'np')
    fn = filter(callable, fn)
    if dtype.lower() in ('pillow', 'pil'):
      getattr(self, target.lower())['transform1'] += list(fn)
    else:
      getattr(self, target.lower())['transform2'] += list(fn)

  def image_augmentation(self):
    """Enable data augmentation

    The data augmentation for single image will randomly rotate and flip the
    image.
    The data augmentation for video will randomly rotate, flip and revert the
    video.

    TODO: revert is not implemented.
    """
    self.aux['augmentation'] = True

  def cropper(self, fn):
    assert callable(fn)
    self.crop = fn

  def set_color_space(self, target: str, mode: str):
    if not mode.upper() in ('RGB', 'L', 'YCbCr', 'Gray'):
      raise ValueError(f"Invalid mode: {mode}, must be RGB | L | YCbCr | Gray")
    assert target.lower() in ('hr', 'lr')
    getattr(self, target.lower()).update(color=mode)

  def make_one_shot_iterator(self, batch_shape, steps, shuffle=None,
                             memory_limit=None):
    """Make an iterator object to generate batch data for models.

    Args:
        batch_shape: The shape of batch to generate.
        steps: The number of batches to generate in one epoch.
        shuffle: A boolean representing whether to shuffle the dataset.
        memory_limit: the maximum system memory to use. (Not GPU memory!!)

    Note:
        The rules for -1 shape:
        - If shape[1] is -1, which represents the temporal length,
          will generate the entire video clips;
        - if shape[2] is -1, will auto choose the channel number according to
          color format;
        - if shape[3] or shape[4] is -1, will deduce the frame height or width
          to replace the value.

        The rules for steps:
        - If the `steps` is -1, will generate batches in sequential order;
        - If the `steps` is an positive integer, the generated batches are
          randomly shuffled.
    """
    shape = list(batch_shape)
    if len(shape) == 4:
      shape.insert(1, 1)
    assert len(shape) is 5
    if shape[-2] != -1 and self.crop is None:
      self.cropper(RandomCrop(self.aux['scale']))
    self.prefetch(shuffle, memory_limit)
    futures.as_completed(self.fs)
    for fs in self.fs:
      if fs.exception():
        raise fs.exception()
      assert fs.done()
    self.fs.clear()
    if not (self.loaded & int(2 ** self.threads - 1)):
      self.data, self.cache = self.cache, self.data
      [self.cache[k].clear() for k in self.cache]
      loaded = self.loaded >> (self.threads * 2)
      if not shuffle:
        loaded += 1  # move to next chunk
        if loaded >= self.aux['cap'] / memory_limit:
          loaded = 0
      self.loaded = loaded << (self.threads * 2)
    return EpochIterator(self, shape, steps, shuffle)

  def prefetch(self, shuffle=None, memory_usage=None):
    # check memory usage
    if isinstance(memory_usage, str):
      memory_usage = Utility.str_to_bytes(memory_usage)
    if not memory_usage:
      memory_usage = FREE_MEMORY
    available_mem = min([np.uint64(memory_usage), np.uint64(FREE_MEMORY)])
    if not self.fs:
      if shuffle:
        self.aux['fetchList'] = list(
            np.random.permutation(self.aux['fetchList']))
      if available_mem > self.aux['cap']:
        LOG.debug("Loading all data into memory.")
        for i in range(self.threads):
          if self.loaded & (1 << i):
            continue
          self.fs.append(self.thp.submit(self._prefetch_all, i))
      else:
        prop = memory_usage / self.aux['cap'] / self.threads
        # How many frames can be read into memory each thread each epoch
        # Note: we assume each "frame" has a close size.
        n = max(1, int(np.round(len(self.hr['data']) * prop)))
        LOG.debug(f"Loading {prop * self.threads * 100:.4f}% data.")
        [self.fs.append(self.thp.submit(self._prefecth_chunk, n, i)) for i in
         range(self.threads)]

  def _prefetch_all(self, index):
    length = len(self.hr['data'])
    # load all clips
    interval = int(np.ceil(length / self.threads))
    frames = []
    names = []
    for img in self.hr['data'][index * interval:(index + 1) * interval]:
      frames.append(img.read_frame(img.frames))
      names.append(img.name)
    self.data['hr'] += frames
    self.data['names'] += names
    if self.hr['data'] is self.lr['data']:
      self.data['lr'] += frames
    else:
      frames = []
      for img in self.lr['data'][index * interval:(index + 1) * interval]:
        frames.append(img.read_frame(img.frames))
      self.data['lr'] += frames
    if self.extra and isinstance(self.extra['data'], Container):
      frames = []
      for img in self.extra['data'][index * interval:(index + 1) * interval]:
        frames.append(img.read_frame(img.frames))
      self.data['extra'] += frames
    self.loaded |= (1 << index)

  def _prefecth_chunk(self, chunk_size, index):
    loaded = self.loaded >> (self.threads * 2)
    n = chunk_size
    st = n * self.threads * loaded  # start chunk
    for i in self.aux['fetchList'][st + n * index:st + n * (index + 1)]:
      img = self.hr['data'][i]
      self.cache['hr'].append(img.read_frame(img.frames))
      self.cache['names'].append(img.name)
      if self.hr['data'] is self.lr['data']:
        self.cache['lr'].append(self.cache['hr'][-1])
      else:
        img = self.lr['data'][i]
        self.cache['lr'].append(img.read_frame(img.frames))
      if self.extra and isinstance(self.extra['data'], Container):
        img = self.extra['data'][i]
        self.cache['extra'].append(img.read_frame(img.frames))
    loaded <<= self.threads
    loaded |= (1 << index)
    self.loaded = loaded << self.threads



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7
from functools import partial


def _exponential_decay(start_lr, steps, decay_step, decay_rate, **kwargs):
  return start_lr * decay_rate ** (steps / decay_step)


def _poly_decay(start_lr, end_lr, steps, decay_step, power, **kwargs):
  return (start_lr - end_lr) * (1 - steps / decay_step) ** power + end_lr


def _stair_decay(start_lr, steps, decay_step, decay_rate, **kwargs):
  return start_lr * decay_rate ** (steps // decay_step)


def _multistep_decay(start_lr, steps, decay_step, decay_rate, **kwargs):
  if not decay_step:
    return start_lr
  for n, s in enumerate(decay_step):
    if steps <= s:
      return start_lr * (decay_rate ** n)
  if steps > decay_step[-1]:
    return start_lr * (decay_rate ** len(decay_step))


def lr_decay(method, lr, **kwargs):
  if method == 'exp':
    return partial(_exponential_decay, start_lr=lr, **kwargs)
  elif method == 'poly':
    return partial(_poly_decay, start_lr=lr, **kwargs)
  elif method == 'stair':
    return partial(_stair_decay, start_lr=lr, **kwargs)
  elif method == 'multistep':
    return partial(_multistep_decay, start_lr=lr, **kwargs)
  else:
    print('invalid decay method!')
    return None



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

# Image customized decoder for NV12([Y][UV/4]), NV21([Y][VU/4])
# NOTE: [Y] means Y channel is a planar channel, [UV] means UV
#   channels together is planar, but U and V are packed. [UV/4]
#   means U and V are sub-sampled by a factor of [2, 2]

import numpy as np
from PIL import ImageFile


class NV12Decoder(ImageFile.PyDecoder):
  """PIL.Image.DECODERS for NV12 format raw bytes

  Registered in `Image.DECODERS`, don't use this class directly!
  """

  def __init__(self, mode, *args):
    super(NV12Decoder, self).__init__(mode, *args)

  def decode(self, buffer):
    if self.mode == 'L':
      # discard UV channel
      self.set_as_raw(buffer, 'L')
    else:
      w, h = self.im.size
      y = np.frombuffer(buffer, 'uint8', count=w * h)
      uv = np.frombuffer(buffer, 'uint8', count=w * h // 2, offset=w * h)
      y = np.reshape(y, [h, w, 1])
      uv = np.reshape(uv, [h // 2, w // 2, 2])
      uv = uv[np.arange(h) // 2][:, np.arange(w) // 2]
      yuv = np.concatenate([y, uv], axis=-1)
      self.set_as_raw(yuv.flatten().tobytes())
    return -1, 0


class NV21Decoder(ImageFile.PyDecoder):
  """PIL.Image.DECODERS for NV21 format raw bytes

  Registered in `Image.DECODERS`, don't use this class directly!
  """

  def __init__(self, mode, *args):
    super(NV21Decoder, self).__init__(mode, *args)

  def decode(self, buffer):
    if self.mode == 'L':
      # discard UV channel
      self.set_as_raw(buffer, 'L')
    else:
      w, h = self.im.size
      y = np.frombuffer(buffer, 'uint8', count=w * h)
      vu = np.frombuffer(buffer, 'uint8', count=w * h // 2, offset=w * h)
      y = np.reshape(y, [h, w, 1])
      vu = np.reshape(vu, [h // 2, w // 2, 2])
      vu = vu[np.arange(h) // 2][:, np.arange(w) // 2]
      uv = vu[:, :, ::-1]
      yuv = np.concatenate([y, uv], axis=-1)
      self.set_as_raw(yuv.flatten().tobytes())
    return -1, 0



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7
import logging
from importlib import import_module

import numpy as np


def open_flo(fn):
  """ Read .flo file in Middlebury format
  # Code adapted from:
  # http://stackoverflow.com/questions/28013200/reading-middlebury-flow-files-with-python-bytes-array-numpy

  # WARNING: this will work on little-endian architectures (eg Intel x86) only!
  # print 'fn = %s'%(fn)
  """
  with open(fn, 'rb') as f:
    magic = np.fromfile(f, np.float32, count=1)
    if 202021.25 != magic:
      logging.error('Magic number incorrect. Invalid .flo file')
      return None
    else:
      w = np.fromfile(f, np.int32, count=1)[0]
      h = np.fromfile(f, np.int32, count=1)[0]
      # print 'Reading %d x %d flo file\n' % (w, h)
      data = np.fromfile(f, np.float32, count=2 * w * h)
      # Reshape data into 3D array (columns, rows, bands)
      # The reshape here is for visualization, the original code is (w,h,2)
      return np.resize(data, (int(h), int(w), 2))


def write_flo(filename, uv, v=None):
  """ Write optical flow to file.

  Original code by Deqing Sun, adapted from Daniel Scharstein.
  """
  n_bands = 2
  _TAG_CHAR = np.array([202021.25], np.float32)

  if v is None:
    u = uv[..., 0]
    v = uv[..., 1]
  else:
    u = uv
  height, width = u.shape
  with open(filename, 'wb') as f:
    # write the header
    f.write(_TAG_CHAR.tobytes())
    np.array(width).astype(np.int32).tofile(f)
    np.array(height).astype(np.int32).tofile(f)
    # arrange into matrix form
    tmp = np.zeros((height, width * n_bands))
    tmp[:, np.arange(width) * 2] = u
    tmp[:, np.arange(width) * 2 + 1] = v
    tmp.astype(np.float32).tofile(f)


class KITTI:
  @staticmethod
  def open_png16(fn):
    """Read 16bit png file"""

    png = import_module('png')
    reader = png.Reader(fn)
    data = reader.asDirect()
    pixels = []
    for row in data[2]:
      row = np.reshape(np.asarray(row), [-1, 3])
      pixels += [row]
    return np.stack(pixels, 0)

  @staticmethod
  def open_flow(fn):
    flow = KITTI.open_png16(fn)
    valid = flow[..., -1]
    u = flow[..., 0].astype('float32')
    v = flow[..., 1].astype('float32')
    u = (u - 2 ** 15) / 64 * valid
    v = (v - 2 ** 15) / 64 * valid
    return np.stack([u, v], -1)



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from .Crop import CenterCrop, RandomCrop
from .Dataset import Container, Dataset, load_datasets
from .Loader import Loader
from .Transform import (
  Bicubic, Brightness, Contrast, FixedVideoLengthBatch, GaussianBlur,
  GaussianWhiteNoise, Sharpness
)

__all__ = [
  'load_datasets',
  'Dataset',
  'Loader',
  'CenterCrop',
  'RandomCrop',
  'Bicubic',
  'Brightness',
  'Contrast',
  'FixedVideoLengthBatch',
  'GaussianWhiteNoise',
  'GaussianBlur',
  'Sharpness'
]



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from PIL import Image, ImageFilter, ImageEnhance
import numpy as np


class Transformer(object):
  """Image transformer.

  Args:
      value: the parameter for each transform function.
      random: if specify 'uniform', generate value sampled from 0 to `+value`;
              if specify 'normal', generate value N~(mean=0, std=value)
  """

  def __init__(self, value=1, random=None):
    self._v = value
    self._r = random

  @property
  def value(self):
    if self._r == 'uniform':
      return np.random.uniform(0, self._v)
    elif self._r == 'normal':
      return np.random.normal(0, self._v)
    else:
      return self._v


class _Transformer1(Transformer):
  def __call__(self, img: Image.Image):
    assert isinstance(img, Image.Image)
    return self.call(img)

  def call(self, img):
    raise NotImplementedError


class Tidy(_Transformer1):
  def call(self, img: Image.Image):
    scale = self.value
    shape = np.array((img.width, img.height))
    shape -= shape % scale
    return img.crop([0, 0, *shape.tolist()])


class Bicubic(_Transformer1):
  def call(self, img: Image.Image):
    scale = self.value
    shape = np.array((img.width, img.height))
    if scale < 1:
      rscale = int(1 / scale)
      if np.any(shape % rscale):
        raise ValueError(f"Image size is not divisible by {rscale}.")
      return img.resize(shape // rscale, resample=Image.BICUBIC)
    else:
      return img.resize((shape * scale).astype('int32'), resample=Image.BICUBIC)


class Brightness(_Transformer1):
  def call(self, img: Image.Image):
    brightness = max(0, self.value)
    return ImageEnhance.Brightness(img).enhance(brightness)


class Contrast(_Transformer1):
  def call(self, img: Image.Image):
    contrast = self.value
    return ImageEnhance.Contrast(img).enhance(contrast)


class Sharpness(_Transformer1):
  def call(self, img):
    sharp = min(max(0, self.value), 2)
    return ImageEnhance.Sharpness(img).enhance(sharp)


class GaussianBlur(_Transformer1):
  def call(self, img):
    radius = self.value
    return ImageFilter.GaussianBlur(radius).filter(img)


class _Transformer2(Transformer):
  def __call__(self, img: np.ndarray):
    assert isinstance(img, np.ndarray)
    return self.call(img)

  def call(self, img):
    raise NotImplementedError


class GaussianWhiteNoise(_Transformer2):
  def call(self, img):
    shape = img.shape
    noise = np.random.normal(0, self.value, shape)
    noise += img.astype('float32')
    return np.clip(np.round(noise), 0, 255).astype('uint8')


class FixedVideoLengthBatch(_Transformer2):
  def call(self, img):
    assert img.ndim == 5
    depth = int(self.value)
    shape = img.shape
    if shape[1] <= depth:
      return img
    ret = []
    for i in range(shape[1] - depth + 1):
      ret.append(img[:, i * depth: (i + 1) * depth])
    return np.stack(ret, 1).reshape([-1, depth, *shape[-3:]])



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from .Config import Config
from .Hook import save_inference_images
from .ImageProcess import (
  array_to_img, img_to_array, imread, imresize, rgb_to_yuv
)
from .LearningRateScheduler import lr_decay
from .Utility import (str_to_bytes, suppress_opt_by_args, to_list, compat_param)


__all__ = [
  'Config',
  'lr_decay',
  'str_to_bytes',
  'suppress_opt_by_args',
  'to_list',
  'array_to_img',
  'imresize',
  'imread',
  'img_to_array',
  'rgb_to_yuv',
  'save_inference_images',
  'compat_param',
]



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import numpy as np
from .ImageProcess import array_to_img
from ..Backend import DATA_FORMAT


def _color_wheel():
  red_yellow, yellow_green, green_cyan = 15, 6, 4
  cyan_blue, blue_magenta, magenta_red = 11, 13, 6
  colors = [red_yellow, yellow_green, green_cyan,
            cyan_blue, blue_magenta, magenta_red]
  color = np.zeros([np.sum(colors), 3])
  for i in range(red_yellow):
    color[i] = [255, 255 * i / red_yellow, 0]
  for i in range(yellow_green):
    color[i + np.sum(colors[:1])] = [255 - 255 * i / yellow_green, 255, 0]
  for i in range(green_cyan):
    color[i + np.sum(colors[:2])] = [0, 255, 255 * i / green_cyan]
  for i in range(cyan_blue):
    color[i + np.sum(colors[:3])] = [0, 255 - 255 * i / cyan_blue, 255]
  for i in range(blue_magenta):
    color[i + np.sum(colors[:4])] = [255 * i / blue_magenta, 0, 255]
  for i in range(magenta_red):
    color[i + np.sum(colors[:5])] = [255, 0, 255 - 255 * i / magenta_red]
  return color / 255


def _viz_flow(u, v, logscale=True, scaledown=6):
  """
  Copied from @jswulff:
    https://github.com/jswulff/pcaflow/blob/master/pcaflow/utils/viz_flow.py

  top_left is zero, u is horizon, v is vertical
  red is 3 o'clock, yellow is 6, light blue is 9, blue/purple is 12
  """
  color_wheel = _color_wheel()
  n_cols = color_wheel.shape[0]

  radius = np.sqrt(u ** 2 + v ** 2)
  if logscale:
    radius = np.log(radius + 1)
  radius = radius / scaledown
  rot = np.arctan2(-v, -u) / np.pi

  fk = (rot + 1) / 2 * (n_cols - 1)  # -1~1 mapped to 0~n_cols
  k0 = fk.astype(np.uint8)  # 0, 1, 2, ..., n_cols

  k1 = k0 + 1
  k1[k1 == n_cols] = 0

  f = fk - k0

  n_colors = color_wheel.shape[1]
  img = np.zeros(u.shape + (n_colors,))
  for i in range(n_colors):
    tmp = color_wheel[:, i]
    col0 = tmp[k0]
    col1 = tmp[k1]
    col = (1 - f) * col0 + f * col1

    idx = radius <= 1
    # increase saturation with radius
    col[idx] = 1 - radius[idx] * (1 - col[idx])
    # out of range
    col[~idx] *= 0.75
    img[:, :, i] = np.floor(255 * col).astype(np.uint8)

  return img.astype(np.uint8)


def visualize_flow(flow, v=None):
  if DATA_FORMAT == 'channels_last':
    u = flow[..., 0] if v is None else flow
    v = flow[..., 1] if v is None else v
  else:
    u = flow[0, :, :] if v is None else flow
    v = flow[1, :, :] if v is None else v
  viz = _viz_flow(u, v)
  return array_to_img(viz, 'RGB')



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from functools import partial
from pathlib import Path

from .ImageProcess import array_to_img


def _str_to_slice(index: str):
  index = index.split(':')
  if len(index) == 1:
    ind = int(index[0])
    if ind < 0:
      sl = slice(ind, None, None)
    else:
      sl = slice(ind, ind + 1)
  else:
    def _maybe_int(x):
      try:
        return int(x)
      except ValueError:
        return None

    sl = slice(*(_maybe_int(i) for i in index))
  return sl


def _save_model_predicted_images(output, names, save_dir, index, auto_rename):
  assert len(names) == 1
  name = names[0]
  for img in output[_str_to_slice(index)]:
    shape = img.shape
    path = Path(save_dir)
    if shape[0] > 1:
      path /= name
    path.mkdir(exist_ok=True, parents=True)
    for i, n in enumerate(img):
      rep = 0
      if auto_rename:
        while (path / f"{name}_id{i:04d}_{rep:04d}.png").exists():
          rep += 1
      path /= f"{name}_id{i:04d}_{rep:04d}.png"
      array_to_img(n).convert('RGB').save(str(path))
  return output


def save_inference_images(save_dir, multi_output_index=-1, auto_rename=None):
  return partial(_save_model_predicted_images, save_dir=save_dir,
                 index=multi_output_index, auto_rename=auto_rename)



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

# Image customized decoder for YV12([Y][U/4][V/4]), YV21([Y][V/4][U/4])
# NOTE: [Y][U][V] means Y/U/V channel is a planar channel, [U/4] means
#   U channel is sub-sampled by a factor of [2, 2]

import numpy as np
from PIL import ImageFile


class YV12Decoder(ImageFile.PyDecoder):
  """PIL.Image.DECODERS for YV12 format raw bytes

  Registered in `Image.DECODERS`, don't use this class directly!
  """

  def __init__(self, mode, *args):
    super(YV12Decoder, self).__init__(mode, *args)

  def decode(self, buffer):
    if self.mode == 'L':
      # discard UV channel
      self.set_as_raw(buffer, 'L')
    else:
      w, h = self.im.size
      y = np.frombuffer(buffer, 'uint8', count=w * h)
      u = np.frombuffer(buffer, 'uint8', count=w * h // 4, offset=w * h)
      v = np.frombuffer(
        buffer, 'uint8', count=w * h // 4, offset=w * h + w * h // 4)
      y = np.reshape(y, [h, w])
      u = np.reshape(u, [h // 2, w // 2])
      v = np.reshape(v, [h // 2, w // 2])
      u = u[np.arange(h) // 2][:, np.arange(w) // 2]
      v = v[np.arange(h) // 2][:, np.arange(w) // 2]
      yuv = np.stack([y, u, v], axis=-1)
      self.set_as_raw(yuv.flatten().tobytes())
    return -1, 0


class YV21Decoder(ImageFile.PyDecoder):
  """PIL.Image.DECODERS for YV21 format raw bytes

  Registered in `Image.DECODERS`, don't use this class directly!
  """

  def __init__(self, mode, *args):
    super(YV21Decoder, self).__init__(mode, *args)

  def decode(self, buffer):
    if self.mode == 'L':
      # discard UV channel
      self.set_as_raw(buffer, 'L')
    else:
      w, h = self.im.size
      y = np.frombuffer(buffer, 'uint8', count=w * h)
      v = np.frombuffer(buffer, 'uint8', count=w * h // 4, offset=w * h)
      u = np.frombuffer(
        buffer, 'uint8', count=w * h // 4, offset=w * h + w * h // 4)
      y = np.reshape(y, [h, w])
      u = np.reshape(u, [h // 2, w // 2])
      v = np.reshape(v, [h // 2, w // 2])
      u = u[np.arange(h) // 2][:, np.arange(w) // 2]
      v = v[np.arange(h) // 2][:, np.arange(w) // 2]
      yuv = np.stack([y, u, v], axis=-1)
      self.set_as_raw(yuv.flatten().tobytes())
    return -1, 0



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import os
import logging
from importlib import import_module
from pathlib import Path

import yaml

try:
  from yaml import FullLoader as _Loader
except ImportError:
  # For older versions
  from yaml import Loader as _Loader

LOG = logging.getLogger('VSR')
HOME = os.environ.get('VSR_HOME')
if not HOME:
  HOME = Path('~').expanduser() / '.vsr'
HOME.mkdir(exist_ok=True, parents=True)
CONFIG = {
  'backend': 'pytorch',
  'verbose': 'info',
}
if Path(HOME / 'config.yml').exists():
  with open(HOME / 'config.yml', encoding='utf8') as fd:
    CONFIG = yaml.load(fd.read(), Loader=_Loader)

LOG.setLevel(CONFIG['verbose'].upper())
hdl = logging.StreamHandler()
hdl.setFormatter(logging.Formatter("%(asctime)s %(levelname)s: %(message)s"))
LOG.addHandler(hdl)

BACKEND = CONFIG['backend'].lower()
if BACKEND == 'auto':
  BACKEND = 'tensorflow'
if BACKEND not in ('tensorflow', 'tensorflow2', 'pytorch'):
  BACKEND = 'pytorch'

if BACKEND in ('tensorflow', 'tensorflow2'):
  try:
    tf = import_module('tensorflow')
    CONFIG['data_format'] = 'channels_last'
    if BACKEND == 'tensorflow2' and tf.__version__.split('.')[0] != '2':
      LOG.warning(f"[!] Current tensorflow version is {tf.__version__}")
      LOG.info("[*] Fallback to use tensorflow")
      BACKEND = 'tensorflow'
  except ImportError:
    LOG.warning("[!] Tensorflow package not found in your system.")
    LOG.info("[*] Fallback to use PyTorch...")
    BACKEND = 'pytorch'

if BACKEND == 'pytorch':
  try:
    torch = import_module('torch')
    CONFIG['data_format'] = 'channels_first'
    _ver = torch.__version__.split('.')
    if _ver[0] != '1' or _ver[1] <= '1':
      LOG.warning(
        f"[!] PyTorch version too low: {torch.__version__}, recommended 1.2.0")
  except ImportError:
    LOG.fatal("[!] PyTorch package not found in your system.")
    raise ImportError("Not an available backend found! Check your environment.")

DATA_FORMAT = CONFIG['data_format'].lower()



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import numpy as np
from ..Backend import DATA_FORMAT


class Cropper(object):
  def __init__(self, scale):
    self.scale = scale

  def __call__(self, img_pair: tuple, shape: list):
    assert len(img_pair) >= 2
    for img in img_pair:
      assert img.ndim == len(shape)

    return self.call(img_pair, shape)

  def call(self, img: tuple, shape: (list, tuple)) -> tuple:
    raise NotImplementedError


class RandomCrop(Cropper):
  def call(self, img: tuple, shape: (list, tuple)) -> tuple:
    hr, lr = img
    if lr.shape[-2] < shape[-2]:
      raise ValueError("Batch shape is too large than data")
    ind = [np.random.randint(nd + 1) for nd in lr.shape - np.array(shape)]
    slc1 = [slice(n, n + s) for n, s in zip(ind, shape)]
    slc2 = slc1.copy()
    if DATA_FORMAT == 'channels_last':
      slc2[-2] = slice(ind[-2] * self.scale,
                       (ind[-2] + shape[-2]) * self.scale)
      slc2[-3] = slice(ind[-3] * self.scale,
                       (ind[-3] + shape[-3]) * self.scale)
    else:
      slc2[-1] = slice(ind[-1] * self.scale,
                       (ind[-1] + shape[-1]) * self.scale)
      slc2[-2] = slice(ind[-2] * self.scale,
                       (ind[-2] + shape[-2]) * self.scale)
    return hr[tuple(slc2)], lr[tuple(slc1)]


class CenterCrop(Cropper):
  def call(self, img: tuple, shape: (list, tuple)) -> tuple:
    hr, lr = img
    ind = [nd // 2 for nd in hr.shape - np.array(shape)]
    slc1 = [slice(n, n + s) for n, s in zip(ind, shape)]
    slc2 = slc1.copy()
    if DATA_FORMAT == 'channels_last':
      slc2[-2] = slice(ind[-2] * self.scale,
                       (ind[-2] + shape[-2]) * self.scale)
      slc2[-3] = slice(ind[-3] * self.scale,
                       (ind[-3] + shape[-3]) * self.scale)
    else:
      slc2[-1] = slice(ind[-1] * self.scale,
                       (ind[-1] + shape[-1]) * self.scale)
      slc2[-2] = slice(ind[-2] * self.scale,
                       (ind[-2] + shape[-2]) * self.scale)
    return hr[tuple(slc2)], lr[tuple(slc1)]



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import numpy as np
from PIL import Image

from ..Backend import DATA_FORMAT


def array_to_img(x: np.ndarray, mode='RGB', min_val=0, max_val=255):
  """Convert an ndarray to PIL Image."""

  x = np.squeeze(x).astype('float32')
  x = (x - min_val) / (max_val - min_val)
  x = x.clip(0, 1) * 255
  if np.ndim(x) == 2:
    return Image.fromarray(x.astype('uint8'), mode='L').convert(mode)
  elif np.ndim(x) == 3:
    if DATA_FORMAT == 'channels_first':
      x = x.transpose([1, 2, 0])
    return Image.fromarray(x.astype('uint8'), mode=mode)
  elif np.ndim(x) == 4:
    if DATA_FORMAT == 'channels_first':
      x = x.transpose([0, 2, 3, 1])
    ret = [Image.fromarray(np.round(i).astype('uint8'), mode=mode) for i in x]
    return ret.pop() if len(ret) is 1 else ret
  elif np.ndim(x) >= 5:
    raise ValueError(f"Dimension of x must <= 4. Got {np.ndim(x)}.")


def img_to_array(img, data_format=None):
  """Converts a PIL Image instance to a Numpy array.

    !!Copy from Keras!!

    Assure the array's ndim is 3
  """
  if not isinstance(img, Image.Image):
    return img
  if data_format is None:
    data_format = DATA_FORMAT
  if data_format not in {'channels_first', 'channels_last'}:
    raise ValueError('Unknown data_format: ', data_format)
  # Numpy array x has format (height, width, channel)
  # or (channel, height, width)
  # but original PIL image1 has format (width, height, channel)
  x = np.asarray(img, dtype=np.uint8)
  if len(x.shape) == 3:
    if data_format == 'channels_first':
      x = x.transpose([2, 0, 1])
  elif len(x.shape) == 2:
    if data_format == 'channels_first':
      x = x.reshape((1, x.shape[0], x.shape[1]))
    else:
      x = x.reshape((x.shape[0], x.shape[1], 1))
  else:
    raise ValueError('Unsupported image1 shape: ', x.shape)
  return x


def imresize(image, scale, size=None, mode=None, resample=None):
  """Image resize using simple cubic provided in PIL"""

  def _resample(name: str):
    if 'cubic' in name:
      return Image.BICUBIC
    if 'linear' in name:
      return Image.BILINEAR
    if 'nearest' in name:
      return Image.NEAREST
    return 0

  dtype = Image.Image
  if isinstance(image, np.ndarray):
    dtype = np.ndarray
    mode = 'RGB' or mode
    image = array_to_img(image, mode)
  if size is None:
    size = (np.array(image.size) * scale).astype(int)
  if image.mode in ('RGB', 'BGR'):
    image = image.convert('YCbCr')
  mode = image.mode if not mode else mode
  if isinstance(resample, str):
    resample = _resample(resample)
  if not resample:
    resample = Image.BICUBIC
  image = image.resize(size, resample=resample).convert(mode)
  if dtype is np.ndarray:
    return img_to_array(image, DATA_FORMAT)
  return image


def imread(url, mode='RGB'):
  """Read image from file to ndarray"""

  img = Image.open(url)
  return img_to_array(img.convert(mode))


_Y601 = (0.299, 0.587, 0.114)
_Y709 = (0.2126, 0.7152, 0.0722)
_UMAX = 0.436
_VMAX = 0.615
_U601 = (_Y601[0] / (_Y601[2] - 1), _Y601[1] / (_Y601[2] - 1), 1.0)
_V601 = (1.0, _Y601[1] / (_Y601[0] - 1), _Y601[2] / (_Y601[0] - 1))
_Y601 = np.array(_Y601, dtype=np.float32)
_U601 = np.array(_U601, dtype=np.float32) * _UMAX
_V601 = np.array(_V601, dtype=np.float32) * _VMAX
_U709 = (_Y709[0] / (_Y709[2] - 1), _Y709[1] / (_Y709[2] - 1), 1.0)
_V709 = (1.0, _Y709[1] / (_Y709[0] - 1), _Y709[2] / (_Y709[0] - 1))
_Y709 = np.array(_Y709, dtype=np.float32)
_U709 = np.array(_U709, dtype=np.float32) * _UMAX
_V709 = np.array(_V709, dtype=np.float32) * _VMAX
_T601 = np.stack([_Y601, _U601, _V601])
_T709 = np.stack([_Y709, _U709, _V709])


def rgb_to_yuv(img, max_val=1.0, standard='bt601'):
  """convert rgb to yuv

  There are plenty of rgb2yuv functions in python modules, but here we want to
  make things more clearly.

  Usually there are two standards: BT.601 and BT.709. While bt601 is the most
  widely used (PIL, opencv, matlab's rgb2gray and also in the lovely
  tf.image), somehow bt709 is not found used in any libs.
  However, matlab's rgb2ycbcr uses different weights, which come from
  C.A. Poynton. Most SR papers use matlab's rgb2ycbcr in benchmark,
  because it gets the highest PSNR :)

  Args:
       img: a 3-D numpy array. If `dtype=uint8`, it ranges from [0, 255], if
         `dtype=float`, it ranges from [0, 1]
       max_val: a scalar representing range of the image value
       standard: a string, should be one of ('bt601', 'bt709', 'matlab')

  Return:
      yuv image
  """
  _standard = standard.lower()
  if _standard not in ('bt601', 'bt709', 'matlab'):
    raise ValueError('Not known standard:', standard)
  if img.shape[-1] != 3:
    return img
  """ matrix used in matlab
    yuv = _trans * rgb + _bias
  """
  _trans = np.array([[65.481, 128.553, 24.966], [-37.797, -74.203, 112],
                     [112, -93.786, -18.214]], dtype=np.float32)
  _bias = np.array([16, 128, 128], dtype=np.float32)
  _trans /= 255
  _bias /= 255
  _img = img.reshape([-1, 3]) / max_val
  if _standard == 'bt601':
    _yuv = np.matmul(_T601, _img.transpose())
    _yuv = _yuv.transpose() + np.array([0, 0.5, 0.5])
    _yuv = _yuv.reshape(img.shape)
  elif _standard == 'bt709':
    _yuv = np.matmul(_T709, _img.transpose())
    _yuv = _yuv.transpose() + np.array([0, 0.5, 0.5])
    _yuv = _yuv.reshape(img.shape)
  else:
    _yuv = np.matmul(_trans, _img.transpose())
    _yuv = _yuv.transpose() + _bias
    _yuv = _yuv.reshape(img.shape)
  return np.clip(_yuv, 0, 1) * max_val



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

from importlib import import_module

from ..Backend import BACKEND

__all__ = [
  'get_model',
  'list_supported_models'
]


def get_model(name: str):
  name = name.lower()
  try:
    if BACKEND == 'pytorch':
      return import_module('.Models', 'VSR.Backend.Torch').get_model(name)
    elif BACKEND == 'tensorflow':
      return import_module('.Models', 'VSR.Backend.TF').get_model(name)
    elif BACKEND == 'tensorflow2':
      pass
  except (KeyError, ImportError):
    raise ImportError(f"Using {BACKEND}, can't find model {name}.")


def list_supported_models():
  if BACKEND == 'pytorch':
    return import_module('.Models', 'VSR.Backend.Torch').list_supported_models()
  elif BACKEND == 'tensorflow':
    return import_module('.Models', 'VSR.Backend.TF').list_supported_models()
  elif BACKEND == 'tensorflow2':
    pass



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import logging
from typing import Generator

from .Config import Config

LOG = logging.getLogger('VSR.Util')


def to_list(x, repeat=1):
  """convert x to list object

    Args:
       x: any object to convert
       repeat: if x is to make as [x], repeat `repeat` elements in the list
  """
  if isinstance(x, (Generator, tuple, set)):
    return list(x)
  elif isinstance(x, list):
    return x
  elif isinstance(x, dict):
    return list(x.values())
  elif x is not None:
    return [x] * repeat
  else:
    return []


def str_to_bytes(s):
  """convert string to byte unit. Case insensitive.

  >>> str_to_bytes('2GB')
    2147483648
  >>> str_to_bytes('1kb')
    1024
  """
  s = s.replace(' ', '')
  if s[-1].isalpha() and s[-2].isalpha():
    _unit = s[-2:].upper()
    _num = s[:-2]
  elif s[-1].isalpha():
    _unit = s[-1].upper()
    _num = s[:-1]
  else:
    return float(s)
  if not _unit in ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'):
    raise ValueError('invalid unit', _unit)
  carry = {
    'B': 1,
    'KB': 1024,
    'MB': 1024 ** 2,
    'GB': 1024 ** 3,
    'TB': 1024 ** 4,
    'PB': 1024 ** 5,
    'EB': 1024 ** 6,
    'ZB': 1024 ** 7,
    'YB': 1024 ** 8
  }
  return float(_num) * carry[_unit]


def cross_type_assign(value, dtype):
  """Convert `value` to `dtype`.
    Usually this can be done by simply `dtype(value)`, however, this ain't
    correct for str -> bool conversion.
  """

  if dtype is bool and isinstance(value, str):
    if value.lower() == 'false':
      return False
    elif value.lower() == 'true':
      return True
    else:
      LOG.warning(
          "suspect wrong typo {}, do you mean true/false?".format(value))
      return True
  return dtype(value)


def suppress_opt_by_args(opt, *args):
  """Use cmdline arguments to overwrite tf declared in yaml file.
    Account for safety, writing section not declared in yaml is not allowed.
  """

  def parse_args(argstr: str, prev_argstr: str):
    if prev_argstr:
      k, v = prev_argstr, argstr
    elif argstr.startswith('--'):
      if '=' in argstr:
        k, v = argstr[2:].split('=')
      else:
        k = argstr[2:]
        v = None
    elif argstr.startswith('-'):
      if '=' in argstr:
        k, v = argstr[1:].split('=')
      else:
        k = argstr[1:]
        v = None
    else:
      raise KeyError("Unknown parameter: {}".format(argstr))
    return k, v

  prev_arg = None
  for arg in args:
    key, value = parse_args(arg, prev_arg)
    prev_arg = None  # clear after use
    if key and value:
      # dict support
      keys = key.split('.')
      if keys[0] not in opt:
        raise KeyError("Parameter {} doesn't exist in model!".format(key))
      old_v = opt.get(keys[0])
      if isinstance(old_v, (list, tuple)):
        # list, tuple support
        if not value.startswith('[') and not value.startswith('('):
          raise TypeError("Invalid list syntax: {}".format(value))
        if not value.endswith(']') and not value.endswith(')'):
          raise TypeError("Invalid list syntax: {}".format(value))
        values = value[1:-1].split(',')
        if len(values) == 1 and values[0] == '':
          # empty list
          values = []
        new_v = [cross_type_assign(nv, type(ov)) for ov, nv in
                 zip(old_v, values)]
        opt[keys[0]] = new_v
      elif isinstance(old_v, dict):
        # dict support
        try:
          for k in keys[1:-1]:
            old_v = old_v[k]
          ref_v = old_v
          old_v = old_v[keys[-1]]
        except KeyError:
          raise KeyError("Parameter {} doesn't exist in model!".format(key))
        if isinstance(old_v, (list, tuple)):
          raise NotImplementedError("Don't support nested list type.")
        new_v = cross_type_assign(value, type(old_v))
        ref_v[keys[-1]] = new_v
      else:
        new_v = cross_type_assign(value, type(old_v))
        opt[keys[0]] = new_v
    elif key:
      prev_arg = key

  if prev_arg:
    raise KeyError("Parameter missing value: {}".format(prev_arg))


def compat_param(par):
  from ..Backend import DATA_FORMAT

  _par = Config(par)
  for i in par:
    if isinstance(par[i], (dict, Config)):
      _par.update(**par[i])
  if 'batch_shape' not in _par:
    assert 'batch' in _par and 'patch_size' in _par and 'channel' in _par
    batch = _par.batch
    channel = _par.channel
    patch_size = _par.patch_size
    patch_size = [x // _par.scale for x in to_list(patch_size, 2)]
    batch_shape = [batch, *patch_size]
    if DATA_FORMAT == 'channels_last':
      batch_shape.append(channel)
    else:
      batch_shape.insert(1, channel)
    if 'depth' in _par and _par.depth != 1:
      batch_shape.insert(1, _par.depth)
    par.batch_shape = batch_shape
    delattr(par, 'batch')
    delattr(par, 'patch_size')
  return par



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import re
from concurrent import futures
from pathlib import Path
import copy

import yaml

from .VirtualFile import ImageFile, RawFile
from ..Util import Config, to_list

try:
  from yaml import FullLoader as _Loader
except ImportError:
  from yaml import Loader as _Loader

IMAGE_SUF = ('PNG', 'JPG', 'JPEG', 'BMP', 'TIFF', 'TIF', 'GIF')
VIDEO_SUF = {
  'NV12': 'NV12',
  'YUV': 'YV12',
  'YV12': 'YV12',
  'NV21': 'NV21',
  'YV21': 'YV21',
  'RGB': 'RGB'
}


def _supported_image(x: Path):
  return x.suffix[1:].upper() in IMAGE_SUF


def _supported_video(x: Path):
  return x.suffix[1:].upper() in VIDEO_SUF


def _supported_suffix(x: Path):
  return _supported_image(x) or _supported_video(x)


class Dataset(object):
  """ Make a `dataset` object

  """

  def __init__(self, *folders):
    self.dirs = list(map(Path, folders))
    self.recursive = True
    self.glob_patterns = ('*',)
    self.inc_patterns = None
    self.exc_patterns = None
    self.as_video = False
    self.compiled = None

  def use_like_video_(self):
    self.as_video = True

  def use_like_video(self):
    d = copy.copy(self)
    d.compiled = None
    d.use_like_video_()
    return d

  def include_(self, *pattern: str):
    self.glob_patterns = list(pattern)
    self.inc_patterns = None

  def include(self, *pattern: str):
    d = copy.copy(self)
    d.compiled = None
    d.include_(*pattern)
    return d

  def include_reg_(self, *reg: str):
    self.inc_patterns = [re.compile(r) for r in reg]
    self.glob_patterns = ('*',)

  def include_reg(self, *reg: str):
    d = copy.copy(self)
    d.compiled = None
    d.include_reg_(*reg)
    return d

  def exclude_(self, *reg: str):
    self.exc_patterns = [re.compile(r) for r in reg]

  def exclude(self, *reg: str):
    d = copy.copy(self)
    d.compiled = None
    d.exclude_(*reg)
    return d

  def compile(self):
    if self.compiled:
      return self.compiled
    files = []

    def _exc(x: Path):
      if self.exc_patterns:
        for reg in self.exc_patterns:
          if reg.search(str(x.absolute().as_posix())):
            return False
      return True

    def _inc(x: Path):
      if self.inc_patterns:
        for reg in self.inc_patterns:
          if reg.search(str(x.absolute().as_posix())):
            return True
      return False

    for folder in self.dirs:
      if not Path(folder).exists():
        continue
      nodes = []
      if folder.is_file():
        # if points to a file rather than a directory
        nodes.append(folder)
      fn_glob = Path.rglob if self.recursive else Path.glob
      for pat in self.glob_patterns:
        nodes += list(fn_glob(folder, pat))
      if self.inc_patterns:
        nodes = filter(_inc, nodes)
      files += list(filter(_exc, filter(_supported_suffix, nodes)))
    image_nodes = list(filter(_supported_image, files))
    if not self.as_video:
      self.compiled = Container(sorted(image_nodes), self.as_video)
      return self.compiled
    video_nodes = list(filter(_supported_video, files))
    video_nodes += list(map(lambda x: x.parent, image_nodes))
    video_nodes = list(set(video_nodes))  # remove duplicated nodes
    self.compiled = Container(sorted(video_nodes), self.as_video)
    return self.compiled


class Container(object):
  """Frames container

  """

  def __init__(self, urls, is_video: bool):
    assert isinstance(urls, (list, tuple))
    pool = futures.ThreadPoolExecutor(4)
    fs = []
    self.nodes = []

    def _parse_image_node(url: Path):
      if url.is_dir():
        for i in filter(_supported_image, url.glob('*')):
          self.nodes.append(ImageFile(i, rewind=True))
      elif _supported_image(url):
        self.nodes.append(ImageFile(url, rewind=True))

    def _parse_video_node(url: Path):
      if _supported_video(url):
        size = re.findall("\\d+x\\d+", url.stem)
        if size:
          size = [int(x) for x in size[0].split('x')]
          self.nodes.append(
              RawFile(url, VIDEO_SUF[url.suffix[1:].upper()], size,
                      rewind=True))
      elif url.is_dir():
        self.nodes.append(ImageFile(url))

    for j in urls:
      if is_video:
        fs.append(pool.submit(_parse_video_node, j))
      else:
        fs.append(pool.submit(_parse_image_node, j))
    futures.as_completed(fs)
    pool.shutdown()

  def __getitem__(self, item):
    return self.nodes[item]

  def __len__(self):
    return len(self.nodes)

  @property
  def capacity(self):
    if not self.nodes:
      return 0
    pos = 0
    max_sz = 0
    total_frames = 0
    for i, n in enumerate(self.nodes):
      total_frames += n.frames
      if n.size() > max_sz:
        max_sz = n.size()
        pos = i
    shape = self.nodes[pos].shape
    max_bpp = 3
    return shape[0] * shape[1] * max_bpp * total_frames


def load_datasets(describe_file, key=''):
  """load dataset described in YAML file"""

  def _extend_pattern(url):
    _url = root / Path(url)
    url_p = _url

    while True:
      try:
        if url_p.exists():
          break
      except OSError:
        url_p = url_p.parent
        continue
      if url_p == url_p.parent:
        break
      url_p = url_p.parent
    # retrieve glob pattern
    url_r = str(_url.relative_to(url_p))
    if url_r == '.' and url_p.is_dir():
      return str(Path(url) / '**/*')
    return url

  def _get_dataset(desc, use_as_video=None, name=None):
    dataset = Config(name=name)
    for i in desc:
      if i not in ('train', 'val', 'test'):
        continue
      if isinstance(desc[i], dict):
        hr = to_list(desc[i].get('hr'))
        lr = to_list(desc[i].get('lr'))
      else:
        hr = to_list(desc[i])
        lr = []
      if use_as_video:
        hr_pattern = [
          x if x not in all_path and x + '[video]' not in all_path else
          all_path[x + '[video]'] for x in hr]
        lr_pattern = [
          x if x not in all_path and x + '[video]' not in all_path else
          all_path[x + '[video]'] for x in lr]
      else:
        hr_pattern = [x if x not in all_path else all_path[x] for x in hr]
        lr_pattern = [x if x not in all_path else all_path[x] for x in lr]
      hr_data = Dataset(root).include(*(_extend_pattern(x) for x in hr_pattern))
      lr_data = Dataset(root).include(
          *(_extend_pattern(x) for x in lr_pattern)) if lr_pattern else None
      hr_data.recursive = False
      if lr_data is not None:
        lr_data.recursive = False
      if use_as_video:
        hr_data.use_like_video_()
        if lr_data is not None:
          lr_data.use_like_video_()
      setattr(dataset, i, Config(hr=hr_data, lr=lr_data))
    return dataset

  datasets = Config()
  with open(describe_file, 'r') as fd:
    config = yaml.load(fd, Loader=_Loader)
    root = Path(config["Root"])
    if not root.is_absolute():
      # make `root` relative to the file
      root = Path(describe_file).resolve().parent / root
      root = root.resolve()
    all_path = config["Path"]
    if key.upper() in config["Dataset"]:
      return _get_dataset(config["Dataset"][key.upper()], name=key)
    elif key.upper() + '[video]' in config["Dataset"]:
      return _get_dataset(config["Dataset"][key.upper() + '[video]'], True,
                          name=key)
    elif key.upper() in all_path:
      return _get_dataset(Config(test=all_path[key.upper()]), name=key)
    elif key.upper() + '[video]' in all_path:
      return _get_dataset(Config(test=all_path[key.upper() + '[video]']), True,
                          name=key)
    for name, value in config["Dataset"].items():
      if '[video]' in name:
        name = name.replace('[video]', '')
        datasets[name] = _get_dataset(value, True, name=name)
      else:
        datasets[name] = _get_dataset(value, name=name)
    for name in all_path:
      if '[video]' in name:
        _name = name.replace('[video]', '')
        datasets[_name] = _get_dataset(Config(test=all_path[name]), True,
                                       name=_name)
      else:
        datasets[name] = _get_dataset(Config(test=all_path[name]), name=name)
    return datasets



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 16

from setuptools import find_packages
from setuptools import setup

# Get version from CHANGELOG
try:
  with open('CHANGELOG.md') as fd:
    VERSION = fd.readline()[:-1]
except IOError:
  VERSION = '0.0.0'

REQUIRED_PACKAGES = [
  'numpy',
  'scipy',
  'scikit-image',
  'matplotlib',
  'pillow',
  'pypng',
  'pytest',
  'PyYAML',
  'psutil',
  'tqdm',
  'h5py',
  'easydict >= 1.9',
]

with open('README.md', 'r', encoding='utf-8') as fd:
  long_desp = fd.read()

setup(
    name='VSR',
    version=VERSION,
    description='Video Super-Resolution Framework',
    long_description=long_desp,
    long_description_content_type="text/markdown",
    url='https://github.com/LoSealL/VideoSuperResolution',
    packages=find_packages(),
    install_requires=REQUIRED_PACKAGES,
    license='MIT',
    author='Wenyi Tang',
    author_email='wenyitang@outlook.com',
    keywords="super-resolution sr vsr cnn srcnn vespcn",
    classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 7

import easydict
import yaml

try:
  from yaml import FullLoader as _Loader
except ImportError:
  from yaml import Loader as _Loader


class Config(easydict.EasyDict):
  def __init__(self, obj=None, **kwargs):
    super(Config, self).__init__(**kwargs)
    if obj is not None:
      assert isinstance(obj, (dict, str))
      if isinstance(obj, str):
        with open(obj, 'r') as fd:
          obj = yaml.load(fd, Loader=_Loader)
      self.update(**obj)

  def __getattr__(self, item):
    return self.get(item)



#  Copyright (c) 2017-2020 Wenyi Tang.
#  Author: Wenyi Tang
#  Email: wenyitang@outlook.com
#  Update: 2020 - 2 - 12

import numpy as np

from .Utility import to_list


def gaussian_kernel(kernel_size: (int, tuple, list), width: float):
  """generate a gaussian kernel

  Args:
      kernel_size: the size of generated gaussian kernel. If is a scalar, the
                   kernel is a square matrix, or it's a kernel of HxW.
      width: the standard deviation of gaussian kernel. If width is 0, the
             kernel is identity, if width reaches to +inf, the kernel becomes
             averaging kernel.
  """

  kernel_size = np.asarray(to_list(kernel_size, 2), np.float)
  half_ksize = (kernel_size - 1) / 2.0
  x, y = np.mgrid[-half_ksize[0]:half_ksize[0] + 1,
         -half_ksize[1]:half_ksize[1] + 1]
  kernel = np.exp(-(x ** 2 + y ** 2) / (2 * width ** 2))
  return kernel / (kernel.sum() + 1e-8)


def anisotropic_gaussian_kernel(kernel_size: (int, tuple, list), theta: float,
                                l1: float, l2: float):
  """generate anisotropic gaussian kernel

  Args:
      kernel_size: the size of generated gaussian kernel. If is a scalar, the
                   kernel is a square matrix, or it's a kernel of HxW.
      theta: rotation angle (rad) of the kernel. [0, pi]
      l1: scaling of eigen values on base 0. [0.1, 10]
      l2: scaling of eigen values on base 1. [0.1, L1]
  """

  def gmdistribution(mu, sigma):
    half_k = (kernel_size - 1) / 2.0
    x, y = np.mgrid[-half_k[0]:half_k[0] + 1, -half_k[1]:half_k[1] + 1]
    X = np.expand_dims(np.stack([y, x], axis=-1), axis=-2)
    L = np.linalg.cholesky(sigma).transpose()
    diag_l = np.diag(L)
    log_det_sigma = 2 * np.log(diag_l).sum()
    log_1h = np.sum(np.matmul((X - mu), np.linalg.inv(L)) ** 2, axis=(-1, -2))
    log_1h = -0.5 * (log_1h + log_det_sigma)
    log_1h -= np.log(2 * np.pi)
    y = np.exp(log_1h)
    return y

  kernel_size = np.array(to_list(kernel_size, 2), np.int)
  theta = np.clip(theta, 0, np.pi)
  l1 = np.clip(l1, 0.1, 10)
  l2 = np.clip(l2, 0.1, l1)
  mat_v = np.array([
    [np.cos(theta), np.sin(theta)],
    [np.sin(theta), -np.cos(theta)]
  ], np.float)
  mat_d = np.array([[l1, 0], [0, l2]], np.float)
  sigma = np.matmul(np.matmul(mat_v, mat_d), np.linalg.inv(mat_v))
  kernel = gmdistribution(0, sigma)
  return kernel / (kernel.sum() + 1e-8)


def list_rshift(l, s):
  for _ in range(s):
    l.insert(0, l.pop(-1))
  return l


def bicubic_filter(x, a=-0.5):
  # https://en.wikipedia.org/wiki/Bicubic_interpolation#Bicubic_convolution_algorithm
  if x < 0:
    x = -x
  if x < 1:
    return ((a + 2.0) * x - (a + 3.0)) * x * x + 1
  if x < 2:
    return (((x - 5) * x + 8) * x - 4) * a
  return 0


def weights_downsample(scale_factor):
  if scale_factor < 1:
    ss = int(1 / scale_factor + 0.5)
  else:
    ss = int(scale_factor + 0.5)
  support = 2 * ss
  ksize = support * 2 + 1
  weights = []
  for lambd in range(ksize):
    dist = -2 + (2 * lambd + 1) / support
    weights.append(bicubic_filter(dist))
  h = np.array([weights])
  h /= h.sum()
  v = h.transpose()
  kernel = np.matmul(v, h)
  assert kernel.shape == (ksize, ksize), f"{kernel.shape} != [{ksize}]"
  return kernel, ss


def weights_upsample(scale_factor):
  if scale_factor < 1:
    ss = int(1 / scale_factor + 0.5)
  else:
    ss = int(scale_factor + 0.5)
  support = 2
  ksize = support * 2 + 1
  weights = [[] for _ in range(ss)]
  for i in range(ss):
    for lambd in range(ksize):
      dist = int((1 + ss + 2 * i) / 2 / ss) + lambd - 1.5 - (2 * i + 1) / 2 / ss
      weights[i].append(bicubic_filter(dist))
  w = [np.array([i]) / np.sum(i) for i in weights]
  w = list_rshift(w, ss - ss // 2)
  kernels = []
  for i in range(len(w)):
    for j in range(len(w)):
      kernels.append(np.matmul(w[i].transpose(), w[j]))
  return kernels, ss


def nd_meshgrid(*size, permute=None):
  _error_msg = ("Permute index must match mesh dimensions, "
                "should have {} indexes but got {}")
  size = np.array(size)
  ranges = []
  for x in size:
    ranges.append(np.linspace(-1, 1, x))
  mesh = np.stack(np.meshgrid(*ranges, indexing='ij'))
  if permute is not None:
    if len(permute) != len(size):
      raise ValueError(_error_msg.format(len(size), len(permute)))
    mesh = mesh[permute]
  return mesh.transpose(*range(1, mesh.ndim), 0)


def camera_response_function(inputs, crf_table, max_val=1):
  """Estimated CRF, transform irradiance L to RGB image. If `crf_table` is
    inverted, transform RGB image to irradiance L.

  Args:
      inputs: A 3-D or 4-D tensor, representing irradiance.
      crf_table: CRF lookup table shape (1024,).
      max_val: specify the range of inputs: in (0, max_val)
  Return:
      RGB images (or L) with the same shape as inputs, in range [0, max_val].
  """

  inputs_norm = np.clip(inputs / max_val, 0, 1)
  quant = crf_table.shape[0] - 1
  inputs_index = (inputs_norm * quant).astype('int32')
  ret = []
  for i in inputs_index.flatten():
    ret.append(crf_table[i])
  return np.reshape(ret, inputs.shape)



"""
Copyright: Wenyi Tang 2017-2018
Author: Wenyi Tang
Email: wenyi.tang@intel.com
Created Date: Dec 21st 2018

Download binary files shared on google drive
"""

import io
import sys
from pathlib import Path

try:
  from googleapiclient.discovery import build
  from googleapiclient.http import MediaIoBaseDownload
  from httplib2 import Http
  from oauth2client import file, client, tools
except ImportError as ex:
  raise ImportError(
    "To download shared google drive file via python,"
    "google-api-python-client, oauth2client is required."
    "Please use pip install google-api-python-client oauth2client.")

SCOPES = 'https://www.googleapis.com/auth/drive.readonly'
CREDENTIALS = './Data/credentials.json'


def require_authorize(store, credentials, scope):
  _argv = sys.argv.copy()
  sys.argv = _argv[:1]
  if '--noauth_local_webserver' in _argv:
    sys.argv.append('--noauth_local_webserver')
  flow = client.flow_from_clientsecrets(credentials, scope)
  creds = tools.run_flow(flow, store)
  sys.argv = _argv
  return creds


def drive_download(name, fileid, path):
  store_path = Path(path) / name
  if store_path.exists():
    print("{} exists, skip download.".format(name))
    return store_path
  # The file token.json stores the user's access and refresh tokens, and is
  # created automatically when the authorization flow completes for the first
  # time.
  store = file.Storage('/tmp/token.json')
  creds = store.get()
  if not creds or creds.invalid:
    creds = require_authorize(store, CREDENTIALS, SCOPES)
  service = build('drive', 'v3', http=creds.authorize(Http()))

  request = service.files().get_media(fileId=fileid)

  fh = io.FileIO(store_path.resolve(), 'wb')
  downloader = MediaIoBaseDownload(fh, request)
  done = False
  while not done:
    status, done = downloader.next_chunk()
    print("\rDownload {}%.".format(int(status.progress() * 100)))
  print('\n', flush=True)
  if done:
    return store_path
